{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* BITMAPS
	 -------
(C) 1998,99 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet.*)

{The Bitmaps unit provides the generic bitmaps and conversion between 1,8,
15,16,24 and 32 bit modes. Windows Bitmaps can be read; they are always
converted to TrueColor to keep palette colors.

In addition the TBitmapCanvas allows to draw on any generic bitmap.}

unit Bitmaps;

interface

uses
	Canvas,CanvasFX;

type
	PGenericBitmap=^TGenericBitmap;
	TGenericBitmap=object(TBitmap)
	private
		LinesPerBlock,BlockCount: Integer;
		Blocks: Pointer;
	public
		procedure Reset(AWidth,AHeight: Integer; AFormat: TBitmapFormat); virtual;
		procedure Convert(AFormat: TBitmapFormat); virtual;
		function EditLine(Y: Integer): Pointer; virtual;
		procedure LoadBMP(const FName: string);
	end;
	PBitmapCanvas=^TBitmapCanvas;
	TBitmapCanvas=object(TCanvas)
		Bitmap: PBitmap;
		constructor Init(ABitmap: PBitmap);
		procedure InitBitmap(ABitmap: PBitmap);
		procedure DirectPutPixel(X,Y: Integer); virtual;
		function DirectGetPixel(X,Y: Integer): TColor; virtual;
	end;

implementation

uses
	GUIUtils,Objects;

type
	PPointerArray=^TPointerArray;
	TPointerArray=array[0..16381] of Pointer;
	T1Array=array[0..65000] of set of 0..7;
	T8Array=array[0..65000] of Byte;
	T1XArray=array[0..32000] of Word;
	T24Array=array[0..21000] of packed record
		R,G,B: Byte;
	end;
	T32Array=array[0..16000] of packed record
		R,G,B,A: Byte;
	end;

procedure TGenericBitmap.Reset(AWidth,AHeight: Integer; AFormat: TBitmapFormat);
var
	I: Integer;
	BlockSize,LastBlockSize: Word;
begin
	if Format=AFormat then
		Exit;
	if (Format<>bfNone) and (AFormat<>bfNone) then
		Reset(0,0,bfNone);
	if AFormat=bfNone then begin
		BlockSize:=LinesPerBlock*LineSize;
		LastBlockSize:=(Height mod LinesPerBlock)*LineSize;
		for I:=0 to BlockCount-2 do
			FreeMem(PPointerArray(Blocks)^[I],BlockSize);
		FreeMem(PPointerArray(Blocks)^[BlockCount-1],LastBlockSize);
		FreeMem(Blocks,BlockCount*SizeOf(Pointer));
		Width:=0;
		Height:=0;
		LineSize:=0;
		Format:=bfNone;
		Exit;
	end else with TColorCode(AFormat) do
		LineSize:=Bytes*AWidth;
	Width:=AWidth;
	Height:=AHeight;
	Format:=AFormat;
	LinesPerBlock:=65528 div LineSize;
	BlockSize:=LinesPerBlock*LineSize;
	LastBlockSize:=(Height mod LinesPerBlock)*LineSize;
	if LastBlockSize=0 then
		LastBlockSize:=BlockSize;
	BlockCount:=Pred(Height+LinesPerBlock) div LinesPerBlock;
	GetMem(Blocks,BlockCount*SizeOf(Pointer));
	for I:=0 to BlockCount-2 do
		GetMem(PPointerArray(Blocks)^[I],BlockSize);
	GetMem(PPointerArray(Blocks)^[BlockCount-1],LastBlockSize);
	ColorKey:=-1;
end;

procedure TGenericBitmap.LoadBMP(const FName: string);
type
	PByteArray=^TByteArray;
	TByteArray=array[0..65528] of Byte;
type
	PRGBArray=^TRGBArray;
	TRGBArray=array[0..20000] of packed record
		R,G,B: Byte;
	end;
var
	BMP: PStream;
	BMPHeader: record
		ID: array[0..1] of Char;
		FileSize: Longint;
		Reserved: array[0..3] of Byte;
		DataOfs,HeaderSize: Longint;
	end;
	BMPInfo: record
		Width,Height: Longint;
		Planes,BPP: Word;
		Compression,ImageSize,HRes,VRes,Colors,IColors: Longint;
		Palette: packed array[Byte] of record
			B,G,R,Reserved: Byte;
		end;
	end;
	AFormat: TBitmapFormat;
	I,Y: Integer;
	Line: PByteArray;
	LSize: Word;
begin
	BMP:=New(PDOSStream,Init(FName,stOpenRead));
	BMP^.Read(BMPHeader,SizeOf(BMPHeader));
	BMP^.Read(BMPInfo,BMPHeader.HeaderSize-4);
	if BMPInfo.BPP<8 then
		RunError;
	if BMPInfo.BPP=8 then
		BMP^.Read(BMPInfo.Palette,1024);
	LSize:=((BMPInfo.BPP+7) shr 3)*BMPInfo.Width;
	Reset(BMPInfo.Width,BMPInfo.Height,bf24);
	if Format<>bfNone then
		for Y:=0 to BMPInfo.Height-1 do begin
			BMP^.Seek(BMPHeader.DataOfs+(BMPInfo.Height-1-Y)*((LSize+3) and (not 3)));
			Line:=EditLine(Y);
			BMP^.Read(Line^,LSize);
			if BMPInfo.BPP=8 then
				for I:=BMPInfo.Width-1 downto 0 do
					Move(BMPInfo.Palette[Line^[I]],PRGBArray(Line)^[I],3);
		end;
	BMP^.Free;
end;

procedure TGenericBitmap.Convert(AFormat: TBitmapFormat);
var
	Converted: PGenericBitmap;
	Pixel,Line: Integer;
	Temp: Word;
	Value: Longint;
begin
	if Format=AFormat then
		Exit;
	New(Converted,Init);
	Converted^.Reset(Width,Height,AFormat);
	Value:=0;
	with BitmapProcessor^ do
		for Line:=0 to Height-1 do
			Convert(ReadLine(Line)^,Format,Converted^.EditLine(Line)^,AFormat,Width);
	Reset(0,0,bfNone);
	Width:=Converted^.Width;
	Height:=Converted^.Height;
	Format:=Converted^.Format;
	LineSize:=Converted^.LineSize;
	InternalID:=Converted^.InternalID;
	Blocks:=Converted^.Blocks;
	LinesPerBlock:=Converted^.LinesPerBlock;
	BlockCount:=Converted^.BlockCount;
	Converted^.Format:=bfNone;
	Converted^.Free;
end;

function TGenericBitmap.EditLine(Y: Integer): Pointer;
var
	WY: Word absolute Y;
begin
	if (Y<0) or (Y>=Height) or (Format=bfNone) then
		RunError;
	EditLine:=Pointer(Longint(PPointerArray(Blocks)^[WY div LinesPerBlock])+(WY mod LinesPerBlock)*LineSize);
end;

constructor TBitmapCanvas.Init(ABitmap: PBitmap);
begin
	inherited Init;
	InitBitmap(ABitmap);
end;

procedure TBitmapCanvas.InitBitmap(ABitmap: PBitmap);
begin
	Bitmap:=ABitmap;
	if Assigned(Bitmap) then
		SurfaceResize(Bitmap^.Width,Bitmap^.Height,False);
	SurfaceClear(True);
end;

procedure TBitmapCanvas.DirectPutPixel(X,Y: Integer);
var
	Dst: Pointer;
	Value: Longint;
	Pixel: Integer absolute X;
begin
	if not Assigned(Bitmap) then
		RunError;
	Value:=GetColor;
	Dst:=Bitmap^.EditLine(Y);
	case Bitmap^.Format of
	bfMono: if Value and $808080>0 then
			Include(T1Array(Dst^)[Pixel shr 3],Pixel and 7)
		else
			Exclude(T1Array(Dst^)[Pixel shr 3],Pixel and 7);
	bf8: T8Array(Dst^)[Pixel]:=RGBToPalette(Value);
	bf15: T1XArray(Dst^)[Pixel]:=((Value shr 9) and $7C00)+((Value shr 6) and $03E0)+((Value shr 3) and $001F);
	bf16: T1XArray(Dst^)[Pixel]:=((Value shr 8) and $F800)+((Value shr 5) and $07E0)+((Value shr 3) and $001F);
	bf24: Move(Value,T24Array(Dst^)[Pixel],3);
	bf32: Longint(T32Array(Dst^)[Pixel]):=Value;
	end;
end;

function TBitmapCanvas.DirectGetPixel(X,Y: Integer): TColor;
var
	Src: Pointer;
	Value: Longint;
	Temp: Word;
	Pixel: Integer absolute X;
begin
	if not Assigned(Bitmap) then
		RunError;
	Src:=Bitmap^.ReadLine(Y);
	case Bitmap^.Format of
	bfMono: Value:=Ord((Pixel and 7) in T1Array(Src^)[Pixel shr 3])*$FFFFFF;
	bf8: Value:=PaletteToRGB(T8Array(Src^)[Pixel]);
	bf15: begin
			Temp:=T1XArray(Src^)[Pixel];
			Value:=Longint(Temp and $7C00) shl 9+(Temp and $03E0) shl 6+(Temp and $001F) shl 3;
		end;
	bf16: begin
			Temp:=T1XArray(Src^)[Pixel];
			Value:=Longint(Temp and $F800) shl 8+(Temp and $07E0) shl 5+(Temp and $001F) shl 3;
		end;
	bf24: Move(T24Array(Src^)[Pixel],Value,3);
	bf32: Value:=Longint(T32Array(Src^)[Pixel]);
	end;
	DirectGetPixel:=Value;
end;

end.