{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* EVENTS
	 ------
(C) 1999 by Arsne von Wyss - asrene@vonwyss.ch

Rev: 19991213

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet. *)

{The Events unit provides very basic messaging mechanisms for objects.
Actually, the only thing it really does it to provide some type definitions
ase basework.

These are the ~TEventObject~ which adds an ~EventHandler~ to the TObject. The
handler itself does not have any function, it has all to be implemented by its
descendants. Events are always stored as ~TEvent~s and can take a Keyboard,
Mouse, Command or Broadcast event. To mark an Event as processed, one can
use ~EventHandled~ which also takes care not to kill broadcast events or
~EventDone~ which will mark the event as handled, no matter of its kind.

To store events, usually before processing them, the ~TEventQueue~ and
~TEventBuider~ can be used. They latter object is just a framework to
implement objects to create new events.}

unit Events;

(* Unit interface is HelpGen enabled *)

interface

uses
	Objects;

const
	{Size of the internal event queue.}
	EventQueueSize=32;
	{Event constants used in the What field of ~TEvent~.}
	evNothing=0;
	evKeyboard=1;
	evKeyDown=evKeyboard+1;
	evKeyUp=evKeyboard+2;
	evMouse=8;
	evMouseMove=evMouse+1;
	evMouseDown=evMouse+2;
	evMouseUp=evMouse+3;
	evMouseDouble=evMouse+4;
	evMouseDrag=evMouse+5;
	evCommand=$10;
	evBroadcast=$20;
	evDesign=$40;

	kbEsc      =$011B;
	kbAltSpace =$0200;
	kbCtrlIns  =$0400;
	kbShiftIns =$0500;
	kbCtrlDel  =$0600;
	kbShiftDel =$0700;
	kbBack     =$0E08;
	kbCtrlBack =$0E7F;
	kbShiftTab =$0F00;
	kbTab      =$0F09;
	kbAltQ     =$1000;
	kbAltW     =$1100;
	kbAltE     =$1200;
	kbAltR     =$1300;
	kbAltT     =$1400;
	kbAltY     =$1500;
	kbAltU     =$1600;
	kbAltI     =$1700;
	kbAltO     =$1800;
	kbAltP     =$1900;
	kbCtrlEnter=$1C0A;
	kbEnter    =$1C0D;
	kbAltA     =$1E00;
	kbAltS     =$1F00;
	kbAltD     =$2000;
	kbAltF     =$2100;
	kbAltG     =$2200;
	kbAltH     =$2300;
	kbAltJ     =$2400;
	kbAltK     =$2500;
	kbAltL     =$2600;
	kbAltZ     =$2C00;
	kbAltX     =$2D00;
	kbAltC     =$2E00;
	kbAltV     =$2F00;
	kbAltB     =$3000;
	kbAltN     =$3100;
	kbAltM     =$3200;
	kbF1       =$3B00;
	kbF2       =$3C00;
	kbF3       =$3D00;
	kbF4       =$3E00;
	kbF5       =$3F00;
	kbF6       =$4000;
	kbF7       =$4100;
	kbF8       =$4200;
	kbF9       =$4300;
	kbF10      =$4400;
	kbHome     =$4700;
	kbUp       =$4800;
	kbPgUp     =$4900;
	kbGrayMinus=$4A2D;
	kbLeft     =$4B00;
	kbRight    =$4D00;
	kbGrayPlus =$4E2B;
	kbEnd      =$4F00;
	kbDown     =$5000;
	kbPgDn     =$5100;
	kbIns      =$5200;
	kbDel      =$5300;
	kbShiftF1  =$5400;
	kbShiftF2  =$5500;
	kbShiftF3  =$5600;
	kbShiftF4  =$5700;
	kbShiftF5  =$5800;
	kbShiftF6  =$5900;
	kbShiftF7  =$5A00;
	kbShiftF8  =$5B00;
	kbShiftF9  =$5C00;
	kbShiftF10 =$5D00;
	kbCtrlF1   =$5E00;
	kbCtrlF2   =$5F00;
	kbCtrlF3   =$6000;
	kbCtrlF4   =$6100;
	kbCtrlF5   =$6200;
	kbCtrlF6   =$6300;
	kbCtrlF7   =$6400;
	kbCtrlF8   =$6500;
	kbCtrlF9   =$6600;
	kbCtrlF10  =$6700;
	kbAltF1    =$6800;
	kbAltF2    =$6900;
	kbAltF3    =$6A00;
	kbAltF4    =$6B00;
	kbAltF5    =$6C00;
	kbAltF6    =$6D00;
	kbAltF7    =$6E00;
	kbAltF8    =$6F00;
	kbAltF9    =$7000;
	kbAltF10   =$7100;
	kbCtrlPrtSc=$7200;
	kbCtrlLeft =$7300;
	kbCtrlRight=$7400;
	kbCtrlEnd  =$7500;
	kbCtrlPgDn =$7600;
	kbCtrlHome =$7700;
	kbAlt1     =$7800;
	kbAlt2     =$7900;
	kbAlt3     =$7A00;
	kbAlt4     =$7B00;
	kbAlt5     =$7C00;
	kbAlt6     =$7D00;
	kbAlt7     =$7E00;
	kbAlt8     =$7F00;
	kbAlt9     =$8000;
	kbAlt0     =$8100;
	kbAltMinus =$8200;
	kbAltEqual =$8300;
	kbCtrlPgUp =$8400;
	kbAltBack  =$0800;
	kbSpace    =$3920;
	kb1        =$0231;
	kb2        =$0332;
	kb3        =$0433;
	kb4        =$0534;
	kb5        =$0635;
	kb6        =$0736;
	kb7        =$0837;
	kb8        =$0938;
	kb9        =$0A39;
	kb0        =$0B30;
	kbQ        =$1071;

	kbRightShift =$01;
	kbLeftShift  =$02;
	kbCtrlShift  =$04;
	kbAltShift   =$08;
	kbScrollState=$10;
	kbNumState   =$20;
	kbCapsState  =$40;
	kbInsState   =$80;

type
	PEventObject=^TEventObject;
	PEventHandle=^TEventHandle;
	PEvent=^TEvent;
	{Record that holds all the information needed for an event. Note that this
	record can carry private data by using one of the InfoXXXX fields.

	When processing events, it's very important to check them for evNothing since
	this means that the event has already been processed.}
	TEvent=record
		Receiver: PEventHandle;
		case What: Word of
		evNothing: (User: Longint);
		evMouse: (Buttons: Byte; Double: ByteBool; Where: TPoint);
		evKeyboard: (case KeyReserved: Word of
			0: (KeyCode: Word);
			1: (CharCode: Char; ScanCode: Byte));
		evCommand: (case Command: Word of
			0: (InfoPtr: Pointer);
			1: (InfoLong: Longint);
			2: (InfoWord: Word);
			3: (InfoInt: Integer);
			4: (InfoByte: Byte);
			5: (InfoChar: Char));
		evDesign: (Message: PEvent);
	end;
	{Queue for events. The ~GUI~'s ~TApplication~ object uses it to store the
	events that were posted but not processed yet.}
	TEventQueue=object(TObject)
		{Number of waiting events. Can be ~EventQueueSize~-1 at highest.}
		Count: Integer;
		{Allocates a new event in the queue and returns the pointer to it.}
		function Put: PEvent;
		{Retrieve an event to be processed. Since this is a FIFO list, this will always be
		the oldest event ~put~ in the queue.}
		function Get: PEvent;
	private
		Head,Tail: Word;
		Event: array[0..Pred(EventQueueSize)] of TEvent;
	end;
	TReferenceObject=object(TObject)
	private
		RefCnt: Integer;
	public
		constructor Init;
		procedure Allocate;
		procedure Free;
		function Unique: Pointer; virtual;
		destructor Done; virtual;
	end;
	TEventHandleObject=object(TReferenceObject)
	public
		function Assigned: Boolean; virtual;
		procedure Event(var AEvent: TEvent); virtual;
		function Unique: Pointer; virtual;
		destructor Done; virtual;
	end;
	TEventHandle=object(TEventHandleObject)
	public
		Owner: PEventObject;
		constructor Init(AOwner: PEventObject);
		procedure Event(var AEvent: TEvent); virtual;
		function Assigned: Boolean; virtual;
	end;
	TEventMultiHandle=object(TEventHandleObject)
	private
		Handles: TCollection;
	public
		constructor Init;
		function Assigned: Boolean; virtual;
		procedure Register(Obj: PEventObject);
		procedure Remove(Obj: PEventObject);
		procedure Event(var AEvent: TEvent); virtual;
		destructor Done; virtual;
	end;
	{This object is about the same fr event handling objects as TObjct is for all the
	others: it provides the very basic features but nor real functionality.}
	TEventObject=object(TObject)
	{HandleEvent will be called with an event to be handled. If the event object
	has event aware children it should resend broadcast events received to them.}
		Handle: PEventHandle;
		constructor Init;
		function AllocateHandle: PEventHandle;
		procedure HandleEvent(var Event: TEvent); virtual;
		destructor Done; virtual;
	end;
	{Main object to build events. It's just a nice interface instead of assigning
	all the event fields, plus it can be used for event queues as well.}
	TEventBuilder=object(TObject)
		{Creates a custom user defined event. The info field cannot be accessed.}
		procedure User(AReceiver: PEventHandle; AWhat: Word; AUser: Longint);
		{Creates a command event togeter with a InfoLong variable.}
		procedure Command(AReceiver: PEventHandle; ACommand: Word; AUser: Longint);
		{Creates a command event togeter with a InfoPtr variable.}
		procedure CommandPtr(AReceiver: PEventHandle; ACommand: Word; AUser: Pointer);
		{Creates a command broadcast event togeter with a InfoLong variable.}
		procedure Broadcast(AReceiver: PEventHandle; ACommand: Word; AUser: Longint);
		{Creates a command broadcast event together with a InfoPtr variable.}
		procedure BroadcastPtr(AReceiver: PEventHandle; ACommand: Word; AUser: Pointer);
		{Opens a new event to be filled with data. You need to override this method when
		implementing your own ~TEventBuilder~ objects. See also ~CloseEvent~.}
		procedure OpenEvent(var Event: PEvent); virtual;
		{Closes an event, this means that all the info has been filled in and the event
		is ready to be processed. One usually overrides this method when implementing
		an own ~TEventBuilder~ object. See also ~OpenEvent~.}
		procedure CloseEvent(Event: PEvent); virtual;
	end;

{Marks the event as processed, no matter what kind of event it is. If you
only want to mark non-broadcast events as processed you should use
~EventHandled~.}
procedure EventDone(var Event: TEvent);

{Marks the event as processed if it isn't a broadvastIf you want to mark
an event as processed without cheking its kind, use ~EventDone~.}
procedure EventHandled(var Event: TEvent);

implementation

function TEventQueue.Put: PEvent;
begin
	Head:=Succ(Head) mod EventQueueSize;
	Put:=@Event[Head];
	if Head=Tail then
		RunError(0); {FULL!!!}
	if Head<>Tail then
		Inc(Count);
end;

function TEventQueue.Get: PEvent;
begin
	if Head=Tail then begin
		Get:=nil;
		Exit;
	end;
	Tail:=Succ(Tail) mod EventQueueSize;
	Get:=@Event[Tail];
	Dec(Count);
end;

procedure EventDone(var Event: TEvent);
begin
	Event.What:=evNothing;
end;

procedure EventHandled(var Event: TEvent);
begin
	with Event do
		if What<>evBroadcast then
			What:=evNothing;
end;

procedure TEventBuilder.User(AReceiver: PEventHandle; AWhat: Word; AUser: Longint);
var
	Event: PEvent;
begin
	if AWhat<>evNothing then begin
		OpenEvent(Event);
		with Event^ do begin
			Receiver:=AReceiver;
			Receiver^.Allocate;
			What:=AWhat;
			User:=AUser;
		end;
		CloseEvent(Event);
	end;
end;

procedure TEventBuilder.Command(AReceiver: PEventHandle; ACommand: Word; AUser: Longint);
var
	Event: PEvent;
begin
	OpenEvent(Event);
	with Event^ do begin
		Receiver:=AReceiver;
		Receiver^.Allocate;
		What:=evCommand;
		Command:=ACommand;
		InfoLong:=AUser;
	end;
	CloseEvent(Event);
end;

procedure TEventBuilder.CommandPtr(AReceiver: PEventHandle; ACommand: Word; AUser: Pointer);
var
	Event: PEvent;
begin
	OpenEvent(Event);
	with Event^ do begin
		Receiver:=AReceiver;
		Receiver^.Allocate;
		What:=evCommand;
		Command:=ACommand;
		InfoPtr:=AUser;
	end;
	CloseEvent(Event);
end;

procedure TEventBuilder.Broadcast(AReceiver: PEventHandle; ACommand: Word; AUser: Longint);
var
	Event: PEvent;
begin
	OpenEvent(Event);
	with Event^ do begin
		Receiver:=AReceiver;
		Receiver^.Allocate;
		What:=evBroadcast;
		Command:=ACommand;
		InfoLong:=AUser;
	end;
	CloseEvent(Event);
end;

procedure TEventBuilder.BroadcastPtr(AReceiver: PEventHandle; ACommand: Word; AUser: Pointer);
var
	Event: PEvent;
begin
	OpenEvent(Event);
	with Event^ do begin
		Receiver:=AReceiver;
		Receiver^.Allocate;
		What:=evBroadcast;
		Command:=ACommand;
		InfoPtr:=AUser;
	end;
	CloseEvent(Event);
end;

procedure TEventBuilder.OpenEvent(var Event: PEvent);
begin
	Abstract;
end;

procedure TEventBuilder.CloseEvent(Event: PEvent);
begin
end;

constructor TReferenceObject.Init;
begin
	inherited Init;
	RefCnt:=1;
end;

procedure TReferenceObject.Allocate;
begin
	Inc(RefCnt);
end;

procedure TReferenceObject.Free;
type
	PReferenceObject=^TReferenceObject;
begin
	Dec(RefCnt);
	if (RefCnt=0) then
		Dispose(PReferenceObject(@self),Done)
end;

function TReferenceObject.Unique: Pointer;
var
	Temp: Pointer;
begin
	if RefCnt>1 then begin
		GetMem(Temp,SizeOf(Self));
		Move(Self,Temp^,SizeOf(Self));
		TReferenceObject(Temp^).RefCnt:=1;
		Dec(RefCnt);
		Unique:=Temp;
	end else
		Unique:=@self;
end;

destructor TReferenceObject.Done;
begin
	if RefCnt>0 then
		Abstract;
end;

function TEventHandleObject.Assigned: Boolean;
begin
	Abstract;
end;

procedure TEventHandleObject.Event(var AEvent: TEvent);
begin
	Abstract;
end;

function TEventHandleObject.Unique: Pointer;
begin
	Abstract;
end;

destructor TEventHandleObject.Done;
begin
	if Assigned then
		Abstract;
	inherited Done;
end;

constructor TEventHandle.Init(AOwner: PEventObject);
begin
	inherited Init;
	Owner:=AOwner;
end;

procedure TEventHandle.Event(var AEvent: TEvent);
begin
	if Assigned then
		Owner^.HandleEvent(AEvent);
end;

function TEventHandle.Assigned: Boolean;
begin
	Assigned:=system.Assigned(Owner);
end;

constructor TEventMultiHandle.Init;
begin
	inherited Init;
	Handles.Init(4,2);
end;

function TEventMultiHandle.Assigned: Boolean;
begin
	Assigned:=Handles.Count>0;
end;

procedure TEventMultiHandle.Register(Obj: PEventObject);
begin
	Handles.Insert(Obj^.AllocateHandle);
end;

procedure TEventMultiHandle.Remove(Obj: PEventObject);
begin
	Handles.Delete(Obj^.Handle);
	Obj^.Handle^.Free;
end;

procedure TEventMultiHandle.Event(var AEvent: TEvent);
var
	I: Integer;
	H: PEventHandle;
begin
	for I:=0 to Handles.Count-1 do begin
		H:=Handles.At(I);
		if not H^.Assigned then begin
			Handles.AtDelete(I);
			H^.Free;
		end else
			H^.Event(AEvent);
	end;
end;

destructor TEventMultiHandle.Done;
var
	I: Integer;
begin
	for I:=0 to Handles.Count-1 do
		PEventHandle(Handles.At(I))^.Free;
	Handles.Done;
	inherited Done;
end;

constructor TEventObject.Init;
begin
	inherited Init;
	New(Handle,Init(@self));
end;

function TEventObject.AllocateHandle: PEventHandle;
begin
	Handle^.Allocate;
	AllocateHandle:=Handle;
end;

procedure TEventObject.HandleEvent(var Event: TEvent);
begin
 {nothing to do...}
end;

destructor TEventObject.Done;
begin
	Handle^.Owner:=nil;
	Handle^.Free;
	inherited Done;
end;

end.