{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

{DRIVER FOR MCGA
 ---------------
(C) 1998,99 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.}

unit Drv_MCGA;

interface

uses
	Objects,GUI,Canvas,CanvasFX,Drv_Gen;

type
{Some proprietary stuff}
	PScreen320x200Data=^TScreen320x200Data;
	TScreen320x200Data=array[0..199,0..319] of Byte;
{Our screen...}
	PScreen320x200=^TScreen320x200;
	TScreen320x200=object(TScreen)
		Data: PScreen320x200Data;
		constructor Init;
		function NewCanvas: PCanvas; virtual;
		destructor Done; virtual;
	end;
{...and its canvas}
	PCanvas320x200=^TCanvas320x200;
	TCanvas320x200=object(TCanvas)
	private
		DidHide: Boolean;
	public
		Color: TColor;
		procedure DirectPutPixel(X,Y: Integer); virtual;
		procedure DirectHLine(X1,Y,X2: Integer); virtual;
		procedure DirectBlit(X1,Y1,X2,Y2,X3,Y3: Integer); virtual;
		function DirectGetPixel(X,Y: Integer): Longint; virtual;
		procedure DirectReadHLine(X1,Y,X2: Integer; var Data); virtual;
		procedure DirectWriteHLine(X1,Y,X2: Integer; const Data); virtual;
		procedure SetColor(AColor: Longint); virtual;
		procedure Lock; virtual;
		procedure UnLock; virtual;
	end;
{Mouse driver}
	PMouseDrv=^TMouseDrv;
	TMouseDrv=object(TMouse)
		constructor Init;
		procedure Show; virtual;
		procedure Hide; virtual;
		procedure Limit(X1,Y1,X2,Y2: Integer); virtual;
		procedure Move(X,Y: Integer); virtual;
		destructor Done; virtual;
	private
		Visible: Integer;
		Back: array[0..15,0..15] of Byte;
	end;
{Driver, replace Open and maybe Check (for Event polling)}
	TDriver=object(TGenericDriver)
		procedure Open; virtual;
		procedure Check; virtual;
	end;

var
	Driver: TDriver;

implementation

uses
	Surfaces,Events,GUIUtils;

var
	MouseEvent: Boolean;
	MouseLock,MK,MX,MY,PX,PY: Integer;

procedure MoveRectPic(X,Y,W,H: Integer; AData: Pointer; ToScreen: Boolean);
begin
	with PScreen320x200(Screen)^ do
		if RectIntersect(X,Y,X+W-1,Y+H-1,0,0,Width-1,Height-1,X,Y,W,H) then begin
			Dec(W,Pred(X));
			if ToScreen then
				for Y:=Y to H do begin
					system.Move(AData^,Data^[Y,X],W);
					Inc(Longint(AData),W);
				end
			else
				for Y:=Y to H do begin
					system.Move(Data^[Y,X],AData^,W);
					Inc(Longint(AData),W);
				end;
		end;
end;

procedure DarkenPixelLine(var Pix: Byte; Count: Word); assembler;
asm
	les di,Pix
	mov dl,6
	mov cx,Count
	cld
	cmp cx,0
	jle @end
@loop:
	mov al,es:[di]
	sub ah,ah
	div dl
	mov bl,ah
	sub ah,ah
	div dl
	mov bh,ah
	shl bh,1
	add bh,ah
	shr bh,2
	mov ah,al
	shl al,1
	add al,ah
	shr al,2
	mov ah,al
	shl bl,1
	add bl,ah
	shr bl,2
	mul dl
	add al,bh
	mul dl
	add al,bl
	stosb
	loop @loop
@end:
end;

procedure DrawMouse;
var
	X,Y,Border: Integer;
	I: Integer;
begin
	with PScreen320x200(Screen)^ do begin
		Border:=Width-PX;
		for Y:=0 to Minimum(15,Height-1-PY) do begin
			DarkenPixelLine(Data^[PY+Y,PX+Y shr 1],Minimum(Border-Y shr 1,Minimum(Y shl 1-Y shr 1,22-Y-Y shr 1)+2));
			DarkenPixelLine(Data^[PY+Y,PX+Y shr 1],Minimum(Border-Y shr 1,Minimum((Y+1) shr 1,22-Y-Y shr 1)+1));
		end;
		Data^[PY+0,PX+0]:=215*((PX xor PY) and 1);
	end;
end;

procedure SetRGBColor(C,R,G,B: Byte); assembler;
asm
	mov dx,$3C8
	mov al,[C]
	out dx,al
	inc dx
	mov al,[R]
	shr al,2
	out dx,al
	mov al,[G]
	shr al,2
	out dx,al
	mov al,[B]
	shr al,2
	out dx,al
end;

{Screen initialization}
constructor TScreen320x200.Init;
var
	I: Integer;
	Col: TRGBColor;
begin
	inherited Init(320,200); {set resolution for clipping}
	asm
		mov ax,$13
		int $10
	end;
	for I:=0 to 255 do begin
		Col.RGB:=PaletteToRGB(I);
		with Col do
			SetRGBColor(I,R,G,B);
	end;
	Data:=Ptr(SegA000,0); {compute data pointer}
	BitmapFormat:=bf8;
end;

function TScreen320x200.NewCanvas: PCanvas;
begin
	NewCanvas:=New(PCanvas320x200,Init);
end;

destructor TScreen320x200.Done;
begin
	asm
		mov ax,$3
		int $10
	end;
	inherited Done;
end;

procedure TCanvas320x200.DirectPutPixel(X,Y: Integer);
begin
	PScreen320x200(Screen)^.Data^[Y,X]:=Color;
end;

procedure TCanvas320x200.DirectBlit(X1,Y1,X2,Y2,X3,Y3: Integer);
var
	I,W: Integer;
begin
	W:=X2-X1+1;
	Dec(Y3,Y1);
	BeginUpdate;
	with PScreen320x200(Screen)^ do
		if Y3<0 then
			for I:=Y1 to Y2 do
				system.Move(Data^[I,X1],Data^[Y3+I,X3],W)
		else
			for I:=Y2 downto Y1 do
				system.Move(Data^[I,X1],Data^[Y3+I,X3],W);
	EndUpdate;
end;

procedure TCanvas320x200.DirectHLine(X1,Y,X2: Integer);
begin
	FillChar(PScreen320x200(Screen)^.Data^[Y,X1],Succ(X2-X1),Color);
end;

function TCanvas320x200.DirectGetPixel(X,Y: Integer): Longint;
begin
	DirectGetPixel:=PScreen320x200(Screen)^.Data^[Y,X];
end;

procedure TCanvas320x200.DirectReadHLine(X1,Y,X2: Integer; var Data);
begin
	system.Move(PScreen320x200(Screen)^.Data^[Y,X1],Data,Succ(X2-X1))
end;

procedure TCanvas320x200.DirectWriteHLine(X1,Y,X2: Integer; const Data);
begin
	system.Move(Data,PScreen320x200(Screen)^.Data^[Y,X1],Succ(X2-X1))
end;

procedure TCanvas320x200.Lock;
begin
	Inc(MouseLock);
	DidHide:=RectOversect(PX-Offset.X,PY-Offset.Y,PX-Offset.X+15,PY-Offset.Y+15);
	if DidHide then
		Mouse^.Hide;
end;

procedure TCanvas320x200.UnLock;
begin
	if DidHide then
		Mouse^.Show
	else if (MX<>PX) or (MY<>PY) then begin
		Mouse^.Hide;
		Mouse^.Show;
	end;
	Dec(MouseLock);
end;

procedure TCanvas320x200.SetColor(AColor: Longint);
begin
	inherited SetColor(AColor);
	Color:=RGBToPalette(AColor);
end;

procedure MouseHandler; far; assembler;
asm
	mov ax,seg @Data
	mov ds,ax
	shr cx,1
	mov MouseEvent,True
	les di,[Mouse]
	cmp es:[di].TMouseDrv.Visible,0
	jne @end
	cmp MouseLock,0
	jne @end
	inc MouseLock
	mov ax,16
	mov si,di
	add si,TMouseDrv.Back
	push cx
	push dx
	push ax
	push ax
	push es
	push si
	push False
	push PX
	push PY
	push ax
	push ax
	push es
	push si
	push True
	mov MK,bx
	mov MX,cx
	mov PX,cx
	mov MY,dx
	mov PY,dx
	call MoveRectPic
	call MoveRectPic
	call DrawMouse
	dec MouseLock
	retf
@end:
	mov MK,bx
	mov MX,cx
	mov MY,dx
end;

constructor TMouseDrv.Init;
begin
	inherited Init;
	Visible:=1;
	asm
		sub	ax,ax
		int	$33
		or 	ax,ax
		jz	@endasm {No mouse?}
		mov ax,$c
		mov cx,$ff
		mov	dx,seg MouseHandler
		mov	es,dx
		mov	dx,offset MouseHandler
		int $33
		mov	ax,3
		int $33
		sub bx,bx
		call MouseHandler
	@endasm:
	end;
end;

procedure TMouseDrv.Show; assembler;
asm
	les di,self
	sub es:[di].Visible,1
	jnz @end
	inc MouseLock
	mov dx,16
	mov ax,MX
	mov PX,ax
	push ax
	mov ax,MY
	mov PY,ax
	push ax
	push dx
	push dx
	push es
	add di,Back
	push di
	push False
	call MoveRectPic
	call DrawMouse
	dec MouseLock
@end:
end;

procedure TMouseDrv.Hide; assembler;
asm
	les di,self
	cmp es:[di].Visible,0
	jne @end
	inc MouseLock
	mov dx,16
	push PX
	push PY
	push dx
	push dx
	push es
	add di,Back
	push di
	push True
	call MoveRectPic
	dec MouseLock
	les di,self
@end:
	inc es:[di].Visible
end;

procedure TMouseDrv.Limit(X1,Y1,X2,Y2: Integer); assembler;
asm
	mov	ax,7
	mov cx,[X1]
	shl cx,1
	mov dx,[X2]
	shl dx,1
	int $33
	mov ax,8
	mov cx,[Y1]
	mov dx,[Y2]
	int $33
	mov	ax,3
	int	$33
	call MouseHandler
end;

procedure TMouseDrv.Move(X,Y: Integer); assembler;
asm
	mov	ax,4
	mov cx,X
	shl cx,1
	mov dx,Y
	int	$33
	mov	ax,3
	int	$33
	call MouseHandler
end;

destructor TMouseDrv.Done;
begin
	asm
		mov ax,$0
		int $33
	end;
	inherited Done;
end;

function TimerDrv: Longint; far; assembler;
asm {Timer tick (18.20648193359/second) -> ms}
	mov es,Seg0040
 db $66; mov ax,es:[$6C]
 db $66; sub dx,dx
 db $66,$0f,$a4,$c2,16 {shld eax,edx,16}
 db $66; shl ax,16
 db $66; sub bx,bx
	mov bx,1193
 db $66; div bx
 db $66; mov dx,ax
 db $66; shr dx,16
end;

procedure TDriver.Open;
begin
	{Initialization order: TIMER -> SCREEN -> MOUSE}
	inherited Open;
	Void(New(PScreen320x200,Init));
	Void(New(PMouseDrv,Init));
end;

procedure TDriver.Check; {Poll for mouse or key events}
begin
	inherited Check;
	if MouseEvent then begin
		MouseEvent:=False;
		Mouse^.Update(MX,MY,MK);
	end;
end;

begin
	Driver.Init;
end.