{DRIVER FOR VESA [Video BIOS Extensions]
 ---------------------------------------
(C) 1998,99 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Reference driver unit for VBE HiColor SVGA with normal mouse/keyboard.

This unit has the follwoing features
------------------------------------
* Runs in both RM and PM
* Supports all 15 and 16 bit VESA banked modes
* Uses directly accessed bitmaps for the mouse cursor
* Smart mouse hiding (only hides when in the canvas' drawing area)
* All the core graphics code is in assembly
* Video Memory access is 4-byte aligned for hlines and bitmaps}

unit Drv_VBE;

interface

uses
	Objects,GUI,Canvas,Bitmaps,Drv_Gen;

type
	{Proprietary types}
	PWordArray=^TWordArray;
	TWordArray=array[0..32766] of Word;
	PVBEInfo=^TVBEInfo;
	TVBEInfo=packed record
		VBESignature: array[0..3] of Char;
		VBEVersion: Word;
		OEMString: PString;
		Capabilities: Longint;
		VideoModes: PWordArray;
		TotalMemory: Word;
		OEMSoftwareRev: Word;
		OEMVendorName,
		OEMProductName,
		OEMProductRev: PString;
		Reserved: array[0..221] of Byte;
		OEMData: array[0..255] of Char;
	end;
	PModeInfo=^TModeInfo;
	TModeInfo=packed record
		{Mandatory information for all VBE versions:}
		ModeAttributes: Word;
		WinAAttributes: Byte;
		WinBAttributes: Byte;
		WinGranularity: Word;
		WinSize: Word;
		WinASegment: Word;
		WinBSegment: Word;
		WinFuncPtr: procedure;
		BytesPerScanLine: Word;
		{Mandatory information for VBE 1.2 and above:}
		Width: Word;
		Height: Word;
		XCharSize: Byte;
		YCharSize: Byte;
		NumberOfPlanes: Byte;
		BitsPerPixel: Byte;
		NumberOfBanks: Byte;
		MemoryModel: Byte;
		BankSize: Byte;
		NumberOfImagePages: Byte;
		_Reserved: Byte;
		{Direct Color fields (required for direct/6 and YUV/7 memory models):}
		RedMaskSize: Byte;
		RedFieldPosition: Byte;
		GreenMaskSize: Byte;
		GreenFieldPosition: Byte;
		BlueMaskSize: Byte;
		BlueFieldPosition: Byte;
		RsvdMaskSize: Byte;
		RsvdFieldPosition: Byte;
		DirectColorModeInfo: Byte;
		{Mandatory information for VBE 2.0 and above:}
		PhysBase,
		OffScreenMemOffset: Longint;
		OffScreenMemSize: Word;
		__Reserved: array[0..205] of Byte;
	end;
	{Our screen...}
	PScreenVESA=^TScreenVESA;
	TScreenVESA=object(TScreen)
	private
		procedure GetVBEInfo;
		function GetModeInfo(Mode: Word; var AModeInfo: TModeInfo): Boolean;
		function SetMode(Mode: Word): Boolean;
		procedure FreeVBEInfo;
	public
		constructor Init;
		function NewCanvas: PCanvas; virtual;
		destructor Done; virtual;
	end;
	{...and its canvas [no DIRECTBAR, DIRECTVLINE and DIRECTBLIT acceleration]}
	PCanvasVESA=^TCanvasVESA;
	TCanvasVESA=object(TCanvas)
	private
		DidHide: Boolean;
	public
		Color: Word;
		procedure DirectPutPixel(X,Y: Integer); virtual;
		procedure DirectHLine(X1,Y,X2: Integer); virtual;
		procedure DirectWriteHLine(X1,Y,X2: Integer; const Data); virtual;
		procedure DirectWriteHLineKey(X1,Y,X2: Integer; const Data; Key: TColor); virtual;
		procedure DirectReadHLine(X1,Y,X2: Integer; var Data); virtual;
		procedure SetColor(AColor: Longint); virtual;
		procedure Lock; virtual;
		procedure UnLock; virtual;
	end;
	{Mouse driver}
	PMouseVESA=^TMouseVESA;
	TMouseVESA=object(TMouse)
	private
		Back: PGenericBitmap;
		Visible,BackX,BackY,Lock: Integer;
		procedure Draw;
		procedure UnDraw;
	public
		Sprite: PGenericBitmap;
		constructor Init;
		procedure Show; virtual;
		procedure Hide; virtual;
		procedure Limit(X1,Y1,X2,Y2: Integer); virtual;
		procedure Move(X,Y: Integer); virtual;
		destructor Done; virtual;
	end;
	{Driver object}
	TDriverVESA=object(TGenericDriver)
		procedure Open; virtual;
		procedure Check; virtual;
	end;

const
	Mode: Word=0{ $114}; {mode $114: 800*600*16}
	Resolution: TPoint=(X:0;Y:0);

var
	Driver: TDriverVESA;
	VBEInfo: TVBEInfo;
	ModeInfo: TModeInfo;

implementation

uses
	Surfaces,Events,GUIUtils,{$IFDEF DPMI}NewFrontier,WinAPI,{$ENDIF}Strings;

{$IFDEF DPMI}
type
	Registers32=packed record
		case Longint of
		0: (EDI,ESI,EBP,ESPRES,EBX,EDX,ECX,EAX:longint;
				Flags,ES,DS,FS,GS,IP,CS,SP,SS: Word);
		1: (DI,DI1,SI,SI1,BP,BP1: Word; _Reserved: Longint;
				BX,B1,DX,D1,CX,C1,AX,A1: Word);
		2: (__Reserved: array[0..3] of Longint;
				BL,BH,B2,B3,DL,DH,D2,D3,CL,CH,C2,C3,AL,AH,A2,A3: Byte);
	end;
	DOSMemPointer=record
		Selector,Segment: Word;
	end;
{$ENDIF}

const
	MouseImg: array[0..15,0..15] of Word=(
	($000,$000,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010),
	($000,$AAA,$000,$000,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010),
	($010,$000,$AAA,$999,$000,$000,$010,$010,$010,$010,$010,$010,$010,$010,$010,$010),
	($010,$000,$BBB,$AAA,$999,$999,$000,$000,$010,$010,$010,$010,$010,$010,$010,$010),
	($010,$010,$000,$BBB,$AAA,$999,$888,$888,$000,$000,$010,$010,$010,$010,$010,$010),
	($010,$010,$000,$BBB,$BBB,$AAA,$999,$888,$777,$777,$000,$000,$010,$010,$010,$010),
	($010,$010,$010,$000,$CCC,$BBB,$AAA,$999,$888,$777,$666,$666,$000,$000,$010,$010),
	($010,$010,$010,$000,$CCC,$CCC,$BBB,$AAA,$999,$888,$777,$666,$555,$555,$000,$000),
	($010,$010,$010,$010,$000,$DDD,$CCC,$BBB,$AAA,$999,$888,$777,$000,$000,$010,$010),
	($010,$010,$010,$010,$000,$DDD,$DDD,$CCC,$BBB,$AAA,$999,$000,$010,$010,$010,$010),
	($010,$010,$010,$010,$010,$000,$EEE,$DDD,$CCC,$BBB,$AAA,$888,$000,$010,$010,$010),
	($010,$010,$010,$010,$010,$000,$EEE,$EEE,$DDD,$000,$CCC,$AAA,$888,$000,$010,$010),
	($010,$010,$010,$010,$010,$010,$000,$FFF,$000,$010,$000,$CCC,$AAA,$888,$000,$010),
	($010,$010,$010,$010,$010,$010,$000,$FFF,$000,$010,$010,$000,$CCC,$AAA,$888,$000),
	($010,$010,$010,$010,$010,$010,$010,$000,$010,$010,$010,$010,$000,$CCC,$000,$010),
	($010,$010,$010,$010,$010,$010,$010,$000,$010,$010,$010,$010,$010,$000,$010,$010));

var
	Bank: Word;
	BankShift: Byte;
	Selector: Word;
	MouseEvent: Boolean;
	MX,MY,MK: Integer;
	SwitchBank: procedure;
	RMask,GMask,BMask: Word;
	RShift,GShift,BShift: Byte;

{$IFDEF DPMI}
function Intr32(IntNo: Byte; var Regs: Registers32): Boolean; assembler;
asm
	mov ax,$300
	sub bx,bx
	mov bl,IntNo
	sub cx,cx
	les di,Regs
	int $31
	cmc
	mov ax,0
	adc ax,ax
end;

function GetRMSelector(Segment: Word): Word; assembler;
asm
	mov ax,2
	mov bx,segment
	int $31
end;
{$ENDIF}

{$IFDEF DPMI}
procedure SwitchBankSingle; far; assembler;
asm
	mov ax,$4F05
	int $10
end;
{$ENDIF}

procedure SwitchBankDouble; far; assembler;
asm
	{$IFDEF DPMI}
	push dx
	mov ax,$4F05
	int $10
	inc bx
	pop dx
	mov ax,$4F05
	int $10
	{$ELSE}
	push dx
	dw $1eff,offset ModeInfo.WinFuncPtr {call far ptr [ModeInfo.WinFuncPtr] -> BP does not assemble this correctly}
	pop dx
	inc bx
	dw $1eff,offset ModeInfo.WinFuncPtr {call far ptr [ModeInfo.WinFuncPtr] -> BP does not assemble this correctly}
	{$ENDIF}
end;


procedure TScreenVESA.GetVBEInfo;
const
	VBESig='VBE2';
var
{$IFDEF DPMI}
	DOSMem: DOSMemPointer;
	Regs: Registers32;
{$ENDIF}
	Temp: Pointer;
	I: Integer;
begin
	{$IFDEF DPMI}
	GetMem(Temp,1024);
	Longint(DOSMem):=GlobalDOSAlloc(SizeOf(TVBEInfo));
	PVBEInfo(Ptr(DOSMem.Selector,0))^.VBESignature:=VBESig;
	FillChar(Regs,SizeOf(Regs),0);
	with Regs do begin
		ES:=DOSMem.Segment;
		AX:=$4F00;
	end;
	Intr32($10,Regs);
	system.Move(Ptr(DOSMem.Selector,0)^,VBEInfo,SizeOf(TVBEInfo));
	asm cld end; {Bugfix for Move32}
	Move32(MapRealPointer(VBEInfo.OEMString),FarPtr(Temp),1024);
	VBEInfo.OEMString:=NewStr(StrPas(Temp));
	asm cld end;
	Move32(MapRealPointer(VBEInfo.OEMVendorName),FarPtr(Temp),1024);
	VBEInfo.OEMVendorName:=NewStr(StrPas(Temp));
	asm cld end;
	Move32(MapRealPointer(VBEInfo.OEMProductName),FarPtr(Temp),1024);
	VBEInfo.OEMProductName:=NewStr(StrPas(Temp));
	asm cld end;
	Move32(MapRealPointer(VBEInfo.OEMProductRev),FarPtr(Temp),1024);
	VBEInfo.OEMProductRev:=NewStr(StrPas(Temp));
	asm cld end;
	Move32(MapRealPointer(VBEInfo.VideoModes),FarPtr(Temp),1024);
	I:=0;
	while PWordArray(Temp)^[I]<>$FFFF do
		Inc(I);
	PWordArray(Temp)^[I]:=0;
	I:=(I+1)*2;
	GetMem(VBEInfo.VideoModes,I);
	system.Move(Temp^,VBEInfo.VideoModes^,I);
	FreeMem(Temp,1024);
	GlobalDOSFree(DOSMem.Selector);
	{$ELSE}
	VBEInfo.VBESignature:=VBESig;
	asm
		mov ax,ds
		mov es,ax
		mov di,offset VBEInfo
		mov ax,$4F00
		int $10
	end;
	VBEInfo.OEMString:=NewStr(StrPas(Pointer(VBEInfo.OEMString)));
	VBEInfo.OEMVendorName:=NewStr(StrPas(Pointer(VBEInfo.OEMVendorName)));
	VBEInfo.OEMProductName:=NewStr(StrPas(Pointer(VBEInfo.OEMProductName)));
	VBEInfo.OEMProductRev:=NewStr(StrPas(Pointer(VBEInfo.OEMProductRev)));
	Temp:=VBEInfo.VideoModes;
	I:=0;
	while PWordArray(Temp)^[I]<>$FFFF do
		Inc(I);
	PWordArray(Temp)^[I]:=0;
	I:=(I+1)*2;
	GetMem(VBEInfo.VideoModes,I);
	system.Move(Temp^,VBEInfo.VideoModes^,I);
	{$ENDIF}
end;

procedure TScreenVESA.FreeVBEInfo;
var
	I: Integer;
begin
	DisposeStr(VBEInfo.OEMString);
	DisposeStr(VBEInfo.OEMVendorName);
	DisposeStr(VBEInfo.OEMProductName);
	DisposeStr(VBEInfo.OEMProductRev);
	I:=0;
	while VBEInfo.VideoModes^[I]<>0 do
		Inc(I);
	FreeMem(VBEInfo.VideoModes,Succ(I)*2);
end;

function TScreenVESA.GetModeInfo(Mode: Word; var AModeInfo: TModeInfo): Boolean;
var
{$IFDEF DPMI}
	DOSMem: DOSMemPointer;
	Regs: Registers32;
{$ELSE}
	Result: Word;
{$ENDIF}
begin
	{$IFDEF DPMI}
	Longint(DOSMem):=GlobalDOSAlloc(SizeOf(TModeInfo));
	FillChar(Regs,SizeOf(Regs),0);
	with Regs do begin
		ES:=DOSMem.Segment;
		EAX:=$4F01;
		ECX:=Mode;
	end;
	Intr32($10,Regs);
	system.Move(Ptr(DOSMem.Selector,0)^,AModeInfo,SizeOf(TModeInfo));
	GlobalDOSFree(DOSMem.Selector);
	GetModeInfo:=Regs.AX=$4F;
	{$ELSE}
	asm
		mov ax,$4F01
		mov cx,[Mode]
		les di,[AModeInfo]
		int $10
		mov [Result],ax
	end;
	GetModeInfo:=Result=$4F;
	{$ENDIF}
end;

function TScreenVESA.SetMode(Mode: Word): Boolean;
var
{$IFDEF DPMI}
	Regs: Registers32;
{$ENDIF}
	Result,I: Word;
begin
	if GetModeInfo(Mode,ModeInfo) then begin
		I:=ModeInfo.WinSize div ModeInfo.WinGranularity;
		BankShift:=0;
		while I shr BankShift>1 do
			Inc(BankShift);
		if (ModeInfo.WinSize<>64) or (I and ((1 shl BankShift)-1)>0) or
				(not Assigned(ModeInfo.WinFuncPtr)) then
			RunError;
		if ModeInfo.WinAAttributes and 6=6 then begin
			{$IFDEF DPMI}
			SwitchBank:=SwitchBankSingle;
			ModeInfo.WinASegment:=GetRMSelector(ModeInfo.WinASegment);
			{$ELSE}
			SwitchBank:=ModeInfo.WinFuncPtr;
			{$ENDIF}
		end else begin
			SwitchBank:=SwitchBankDouble;
			{$IFDEF DPMI}
			ModeInfo.WinASegment:=GetRMSelector(ModeInfo.WinASegment);
			ModeInfo.WinBSegment:=GetRMSelector(ModeInfo.WinBSegment);
			{$ENDIF}
		end;
		Bank:=0;
		BMask:=($FFFF shr (16-ModeInfo.BlueMaskSize)) shl ModeInfo.BlueFieldPosition;
		GMask:=($FFFF shr (16-ModeInfo.GreenMaskSize)) shl ModeInfo.GreenFieldPosition;
		RMask:=($FFFF shr (16-ModeInfo.RedMaskSize)) shl ModeInfo.RedFieldPosition;
		BShift:=(8-ModeInfo.BlueMaskSize)-ModeInfo.BlueFieldPosition;
		GShift:=(16-BShift-ModeInfo.GreenMaskSize)-ModeInfo.GreenFieldPosition;
		RShift:=(24-BShift-GShift-ModeInfo.RedMaskSize)-ModeInfo.RedFieldPosition;
		BitmapFormat:=ModeInfo.RedMaskSize shl 8+ModeInfo.GreenMaskSize shl 4+ModeInfo.BlueMaskSize;
		{$IFDEF DPMI}
		FillChar(Regs,SizeOf(Regs),0);
		with Regs do begin
			AX:=$4F02;
			BX:=Mode;
		end;
		Intr32($10,Regs);
		Result:=Regs.AX;
		{$ELSE}
		asm
			mov ax,$4F02
			mov bx,[Mode]
			int $10
			mov [Result],ax
		end;
		{$ENDIF}
		if Result=$4F then
			SetMode:=True
		else begin
			SetMode:=False;
			WriteLn('SetMode failed - video mode error');
		end;
	end else begin
		SetMode:=False;
		WriteLn('SetMode failed - mode info error');
	end;
end;

{Screen initialization}
constructor TScreenVESA.Init;
var
	ModeKey: record
		Count: Integer;
		Data: array[0..60] of Word;
	end;
	I: Integer;
	M: Word;
begin
	inherited Init(0,0);
	GetVBEInfo;
	if VBEInfo.VBESignature<>'VESA' then
		Halt(1);
	if Mode=0 then begin
		I:=0;
		if Longint(Resolution)<>0 then begin
			repeat
				M:=VBEInfo.VideoModes^[I];
				if M=0 then
					Break;
				Inc(I);
				GetModeInfo(M,ModeInfo);
				if (ModeInfo.BitsPerPixel in [9..16]) and (ModeInfo.Width=Resolution.X) and (ModeInfo.Height=Resolution.Y) then begin
					Mode:=M;
					if ModeInfo.BitsPerPixel=16 then
						Break;
				end;
			until False;
		end else begin
			WriteLn('VBE detected: '+VBEInfo.OEMString^);
			Writeln('Select mode');
			WriteLn('');
			WriteLn('Nr Resolution Colors');
			ModeKey.Count:=0;
			repeat
				M:=VBEInfo.VideoModes^[I];
				if M=0 then
					Break;
				Inc(I);
				GetModeInfo(M,ModeInfo);
				if ModeInfo.BitsPerPixel in [9..16] then begin
					WriteLn(ModeKey.Count:2,': ',ModeInfo.Width:4,'*',ModeInfo.Height:4,' ',(Longint(1) shl ModeInfo.BitsPerPixel):5);
					ModeKey.Data[ModeKey.Count]:=M;
					Inc(ModeKey.Count);
				end;
			until False;
			ReadLn(I);
			if (I<0) or (I>=ModeKey.Count) then
				Halt(1);
			Mode:=ModeKey.Data[I];
		end;
	end;
	if (Mode<>0) and GetModeInfo(Mode,ModeInfo) then begin
		Resize(ModeInfo.Width,ModeInfo.Height);
		if SetMode(Mode) then
			Exit;
	end;
	Halt(1);
end;

function TScreenVESA.NewCanvas: PCanvas;
begin
	NewCanvas:=New(PCanvasVESA,Init);
end;

destructor TScreenVESA.Done;
begin
	asm
		mov ax,$3
		int $10
	end;
	FreeVBEInfo;
	inherited Done;
end;

procedure TCanvasVESA.DirectPutPixel(X,Y: Integer); assembler;
asm
	mov ax,[Y]
	mul [ModeInfo.BytesPerScanline]
	mov si,[X]
	shl si,1
	add si,ax
	adc dx,0
	mov cl,[BankShift]
	shl dx,cl
	cmp dx,[Bank]
	je @BankOk
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
@BankOk:
	les di,[self]
	mov ax,word ptr es:[di].Color
	mov es,[ModeInfo.WinASegment]
	mov es:[si],ax
end;

procedure TCanvasVESA.DirectHLine(X1,Y,X2: Integer); assembler;
asm
	cld
	les si,[self]
	mov ax,[Y]
	mul [ModeInfo.BytesPerScanline]
	mov di,[X1]
	shl di,1
	add di,ax
	adc dx,0
	mov cl,[BankShift]
	shl dx,cl
	cmp dx,[Bank]
	je @BankOk
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
@BankOk:
	mov cx,[X2]
	sub cx,[X1]
	inc cx
	mov ax,es:[si].Color
	mov [Y],ax
	mov es,[ModeInfo.WinASegment]
	mov ax,$FFFF
	sub ax,di
	add ax,1
	jc @NoSwitch {if $10000}
	shr ax,1
	cmp ax,cx
	jae @NoSwitch
@Switch:
	sub cx,ax
	mov bx,cx
	mov cx,ax
	mov ax,[Y]
	test di,2
	jz @Align1
	stosw
	dec cx
	jz @EndCopy
@Align1:
	shr cx,1
	db $66; shl ax,16
	mov ax,[Y]
	db $66; rep stosw
@EndCopy:
	push bx
	mov dx,[Bank]
	inc dx
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
	pop cx
@NoSwitch:
	mov ax,[Y]
	test di,2
	jz @Align2
	stosw
	dec cx
	jz @End
@Align2:
	db $66; shl ax,16
	mov ax,[Y]
	shr cx,1
	jz @NoLong
	db $66; rep stosw
	jnc @End
@NoLong:
	stosw
@End:
end;

procedure TCanvasVESA.DirectWriteHLine(X1,Y,X2: Integer; const Data); assembler;
asm
	push ds
	cld
	mov ax,[Y]
	mul [ModeInfo.BytesPerScanline]
	mov di,[X1]
	shl di,1
	add di,ax
	adc dx,0
	mov cl,[BankShift]
	shl dx,cl
	cmp dx,[Bank]
	je @BankOk
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
@BankOk:
	mov cx,[X2]
	sub cx,[X1]
	inc cx
	mov es,[ModeInfo.WinASegment]
	lds si,[Data]
	mov ax,$FFFF
	sub ax,di
	add ax,1
	jc @NoSwitch {if $10000}
	shr ax,1
	cmp ax,cx
	jae @NoSwitch
@Switch:
	sub cx,ax
	mov bx,cx
	mov cx,ax
	test di,2
	jz @Align1
	movsw
	dec cx
	jz @EndCopy
@Align1:
	shr cx,1
	db $66; rep movsw
@EndCopy:
	mov ax,ds
	pop ds
	push ax
	push bx
	mov dx,[Bank]
	inc dx
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
	pop cx
	mov ax,ds
	pop ds
	push ax
@NoSwitch:
	test di,2
	jz @Align2
	movsw
	dec cx
	jz @End
@Align2:
	shr cx,1
	jz @NoLong
	db $66; rep movsw
	jnc @End
@NoLong:
	movsw
@End:
	pop ds
end;

procedure TCanvasVESA.DirectWriteHLineKey(X1,Y,X2: Integer; const Data; Key: TColor); assembler;
asm
	push ds
	cld
	mov ax,[Y]
	mul [ModeInfo.BytesPerScanline]
	mov di,[X1]
	shl di,1
	add di,ax
	adc dx,0
	mov cl,[BankShift]
	shl dx,cl
	cmp dx,[Bank]
	je @BankOk
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
@BankOk:
	mov cx,[X2]
	sub cx,[X1]
	inc cx
	mov es,[ModeInfo.WinASegment]
	lds si,[Data]
	mov ax,$FFFF
	sub ax,di
	add ax,1
	jc @NoSwitch {if $10000}
	shr ax,1
	cmp ax,cx
	jae @NoSwitch
@Switch:
	sub cx,ax
	push cx
	mov bx,word ptr Key
	jmp @LoopBegin
@LoopStart:
	add di,2 {carry on end!}
	jc @LoopEnd
@LoopBegin:
	lodsw
	cmp ax,bx
	je @LoopStart
	mov es:[di],ax
	jmp @LoopStart
@LoopEnd:
	mov bx,sp
	mov ax,ds
	xchg ax,ss:[bx+2]
	mov ds,ax
	mov dx,[Bank]
	inc dx
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
	pop cx
	mov ax,ds
	pop ds
	push ax
@NoSwitch:
	mov bx,word ptr Key
	jmp @Loop2Begin
@Loop2Start:
	add di,2
	dec cx
	jz @End
@Loop2Begin:
	lodsw
	cmp ax,bx
	je @Loop2Start
	mov es:[di],ax
	jmp @Loop2Start
@End:
	pop ds
end;

procedure TCanvasVESA.DirectReadHLine(X1,Y,X2: Integer; var Data); assembler;
asm
	push ds
	cld
	mov ax,[Y]
	mul [ModeInfo.BytesPerScanline]
	mov si,[X1]
	shl si,1
	add si,ax
	adc dx,0
	mov cl,[BankShift]
	shl dx,cl
	cmp dx,[Bank]
	je @BankOk
	mov [Bank],dx
	sub bx,bx
	call SwitchBank
@BankOk:
	mov cx,[X2]
	sub cx,[X1]
	inc cx
	mov ds,[ModeInfo.WinASegment]
	les di,[Data]
	mov ax,$FFFF
	sub ax,si
	add ax,1
	jc @NoSwitch {if $10000}
	shr ax,1
	cmp ax,cx
	jae @NoSwitch
@Switch:
	sub cx,ax
	mov bx,cx
	mov cx,ax
	test si,2
	jz @Align1
	movsw
	dec cx
	jz @EndCopy
@Align1:
	shr cx,1
	db $66; rep movsw
@EndCopy:
	mov ax,ds
	pop ds
	push ax
	mov dx,[Bank]
	inc dx
	mov [Bank],dx
	push bx
	sub bx,bx
	call SwitchBank
	pop cx
	mov ax,ds
	pop ds
	push ax
@NoSwitch:
	test si,2
	jz @Align2
	movsw
	dec cx
	jz @End
@Align2:
	shr cx,1
	jz @NoLong
	db $66; rep movsw
	jnc @End
@NoLong:
	movsw
@End:
	pop ds
end;

procedure TCanvasVESA.Lock;
var
	PX,PY: Integer;
begin
	with PMouseVESA(Mouse)^ do begin
		Inc(Lock);
		PX:=BackX-Offset.X;
		PY:=BackY-Offset.Y;
		DidHide:=(PMouseVESA(Mouse)^.Visible>0) or RectOversect(PX,PY,Pred(PX+Sprite^.Width),Pred(PY+Sprite^.Height));
		if DidHide then
			Hide;
	end;
end;

procedure TCanvasVESA.UnLock;
begin
	with PMouseVESA(Mouse)^ do begin
		if DidHide then
			Show
		else if ((BackX<>MX) or (BackY<>MY)) and (Visible=0) then begin
			UnDraw;
			Draw;
		end;
		Dec(Lock);
	end;
end;

procedure TCanvasVESA.SetColor(AColor: Longint); assembler;
asm
	db $66; push word ptr AColor
	db $66; push word ptr [self]
	call TCanvas.SetColor
	db $66; mov ax,word ptr AColor
	mov cl,BShift
	db $66; shr ax,cl
	mov bx,ax
	and bx,BMask
	mov cl,GShift
	db $66; shr ax,cl
	mov dx,ax
	and dx,GMask
	mov cl,RShift
	db $66; shr ax,cl
	and ax,RMask
	or ax,bx
	or ax,dx
	les di,[self]
	mov es:[di].Color,ax
end;

procedure MouseHandler; far; assembler;
asm
	mov ax,seg @Data
	mov ds,ax
	mov MouseEvent,True
	mov ax,MK
	sub ax,bx
	mov MK,bx
	mov MX,cx
	mov MY,dx
	cmp ax,0
	je @end
	push cx
	push dx
	push bx
	db $66; push word ptr Mouse
	call TMouseVESA.Update
	mov MouseEvent,False
@end:
	les di,[Mouse]
	cmp es:[di+TMouseVESA.Visible],0
	jne @end2
	cmp es:[di+TMouseVESA.Lock],0
	jne @end2
	inc es:[di+TMouseVESA.Lock]
	push es
	push di
	push es
	push di
	call TMouseVesa.UnDraw
	call TMouseVesa.Draw
	les di,[Mouse]
	dec es:[di+TMouseVESA.Lock]
@end2:
end;

constructor TMouseVESA.Init;
var
	X,Y: Integer;
	W: Word;
begin
	inherited Init;
	Visible:=1;
	New(Back,Init);
	New(Sprite,Init);
	Sprite^.Reset(16,16,Screen^.BitmapFormat);
	Sprite^.ColorKey:=32;
	with PBitmapCanvas(New(PBitmapCanvas,Init(Sprite)))^ do begin
		for Y:=0 to Height-1 do
			for X:=0 to Width-1 do begin
				W:=MouseImg[Y,X];
				if W=$010 then
					SetColor(1 shl (16-ModeInfo.GreenMaskSize))
				else if W=0 then
					SetColor(clBlack)
				else
					SetColor(Longint(W and $F00) shl 12+(W and $F0) shl 8+(W and $F) shl 4 or $0F0F0F);
				PutPixel(X,Y);
			end;
		Free;
	end;
	asm
		sub	ax,ax
		int	$33
		or 	ax,ax
		jz	@endasm {No mouse?}
		mov ax,$c
		mov cx,$ff
		mov	dx,seg MouseHandler
		mov	es,dx
		mov	dx,offset MouseHandler
		int $33
		mov	ax,3
		int $33
		sub bx,bx
		call MouseHandler
	@endasm:
	end;
	Limit(0,0,Screen^.Width-1,Screen^.Height-1);
	Move(Screen^.Width shr 1,Screen^.Height shr 1);
	MK:=0;
end;

procedure TMouseVESA.Show; assembler;
asm
	les di,self
	dec es:[di].Visible
	jnz @end
	inc es:[di+TMouseVESA.Lock]
	push es
	push di
	call Draw
	les di,self
	dec es:[di+TMouseVESA.Lock]
@end:
end;

procedure TMouseVESA.Hide; assembler;
asm
	les di,self
	cmp es:[di].Visible,0
	jnz @end
	inc es:[di+TMouseVESA.Lock]
	push es
	push di
	call UnDraw
	les di,self
	dec es:[di+TMouseVESA.Lock]
@end:
	inc es:[di].Visible
end;

procedure TMouseVESA.Limit(X1,Y1,X2,Y2: Integer); assembler;
asm
	mov	ax,7
	mov cx,[X1]
	mov dx,[X2]
	int $33
	mov ax,8
	mov cx,[Y1]
	mov dx,[Y2]
	int $33
	mov	ax,3
	int	$33
	call MouseHandler
end;

procedure TMouseVESA.Move(X,Y: Integer); assembler;
asm
	mov	ax,4
	mov cx,X
	mov dx,Y
	int	$33
	mov	ax,3
	int	$33
	call MouseHandler
end;

procedure TMouseVESA.Draw;
var
	X1,Y1,X2,Y2,Y: Integer;
begin
	BackX:=MX;
	BackY:=MY;
	if Assigned(Sprite) and RectIntersect(BackX,BackY,Pred(BackX+Sprite^.Width),Pred(BackY+Sprite^.Height),
				0,0,Screen^.Width-1,Screen^.Height-1,X1,Y1,X2,Y2) then begin
		if (Back^.Width<>Sprite^.Width) or (Back^.Height<>Sprite^.Height) then
			Back^.Reset(Sprite^.Width,Sprite^.Height,Screen^.BitmapFormat);
		with Screen^.Canvas^ do
			for Y:=Y1 to Y2 do begin
				DirectReadHLine(X1,Y,X2,Back^.EditLine(Y-Y1)^);
				DirectWriteHLineKey(X1,Y,X2,Sprite^.ReadLine(Y-Y1)^,Sprite^.ColorKey);
			end;
	end else
		Back^.Reset(0,0,bfNone);
end;

procedure TMouseVESA.UnDraw;
var
	X1,Y1,X2,Y2,Y: Integer;
begin
	if RectIntersect(BackX,BackY,Pred(BackX+Back^.Width),Pred(BackY+Back^.Height),
				0,0,Screen^.Width-1,Screen^.Height-1,X1,Y1,X2,Y2) and (Back^.Width>0) and (Back^.Height>0) then
		with Screen^.Canvas^ do
			for Y:=Y1 to Y2 do
				DirectWriteHLine(X1,Y,X2,Back^.ReadLine(Y-Y1)^);
end;

destructor TMouseVESA.Done;
begin
	asm
		mov ax,$0
		int $33
	end;
	Back^.Free;
	Sprite^.Free;
	inherited Done;
end;

procedure TDriverVESA.Open;
begin
	{Initialization order: SCREEN -> MOUSE}
	inherited Open;
	Void(New(PScreenVESA,Init));
	Void(New(PMouseVESA,Init));
end;

procedure TDriverVESA.Check; {Poll for mouse or key events}
begin
	inherited Check;
	if MouseEvent then begin
		MouseEvent:=False;
		Mouse^.Update(MX,MY,MK);
	end;
end;

begin
	Driver.Init;
end.
